# frozen_string_literal: true

module Dcc
  class Packet
    attr_reader :address, :address_type, :data, :error_check

    def initialize(address_type, address, data, error_check = nil)
    end
    private_class_method :new

    def self.from_bytes(*bytes)
      unless bytes.all? { |byte| byte.between? 0, 255 }
        fail ArgumentError "#{bytes.inspect} contains an out of range byte"
      end

      @bytes = bytes.clone
      @error_check = bytes.last
      if bytes.first.zero?
        @address_type = :loco_broadcast
        @address = bytes[0]
        @data = bytes[1..-2]
      elsif bytes.first == 255
        @address_type = :idle
        @address = bytes[0]
        @data = bytes[1..-2]
      elsif bytes.first.between? 1, 127
        @address_type = :short_loco
        @address = bytes[0]
        @data = bytes[2..-2]
      elsif bytes.first.between? 128, 191
        # Accessory decoder
        if byte[1] < 128
          # 11 bit (signalling)
          @address_type = :signal
          @address = ((bytes[1] & B00000110) << 7) + ((bytes[1] & B01110000) << 4) + (bytes[0] & B00111111)
          @address_type = :signal_broadcast if @address == 831
          @data = bytes[2..-2]
        else
          # 9 bit
          @address_type = :accessory
          @address = ((~bytes[1] & B01110000) << 2) + (bytes[0] & B00111111)
          @address_type = :accessory_broadcast if @address == 511
          @data = bytes[2..-2]
          @data[0] &= B00001111
        end
      elsif bytes.first.between? 192, 231
        @address_type = :long_loco
        @address = long_loco_address bytes[0], bytes[1]
        @data = bytes[2..-2]
      else
        @address_type = :reserved
        @address = bytes[0]
        @data = bytes[1..-2]
      end
    end

    def to_bytes
      @bytes.clone
    end

    def addressed_to?(type, value, match_broadcast: true, match_idle: false)
      (address_type == type && address == value) ||
        (match_broadcast && (type == :short_loco || type == :long_loco) && address_type == :loco_broadcast) ||
        (match_broadcast && type == :accessory && address_type == :accessory_broadcast) ||
        (match_broadcast && type == :signal && address_type == :signal_broadcast) ||
        (match_idle && address_type == :idle)
    end
  end
end
